module PPrint(pretty,prettySourceFile,
              prettyStr,prettyType,prettyTheo,prettyExp,prettyDef,prettyDlist,
              prettyCeq,prettyEqn,
              ppProg,ppExp,ppEqn,ppType,allTypes,ppSourceFile,
              -- UU_Pretty 2 compatibility
               display, wdisplay, pp_sp, pp_char, spaces
             ) 
where

import UU_Pretty 
import UU_Pretty_ext
import OldPretty

import Syntax
import Expressions 
import Constants
import Types

-- { UU_Pretty 2 compatibility


--   UU_Pretty 2 compatibility }

-- { Extended combinators

hv_sep :: (PP a, PP b) => b -> b -> [a] -> PP_Doc
hv_sep _     _   []      = empty
hv_sep opsep _   [d]     = opsep >|< pp d
hv_sep opsep sep (d:ds)  = lhv_sep ((opsep >|< pp d) : map ((sep >|<) . pp) ds)

lhv_sep fs@(f:fss) = hs >>^<< vs >>$< fs
  where (hs, vs)  = foldr paralg (par, par) fss
        paralg    = \_ (nhs,nvs) -> (eelement_h1 par >>|<< nhs, par >>-<< nvs)

--   Extended combinators }

pretty    p = render  (ppProg p) page_width
prettySourceFile  = display .   ppSourceFile
prettyStr   = display . ppProg
prettyType  = display . ppType
prettyTheo  = display . ppTheor
prettyExp   = display . ppExp 0
prettyDef   = display . ppDef
prettyDlist = display . vlist . ppDlist
prettyCeq   = display . ppCeq
prettyEqn   = display . ppEqn

ppSourceFile (SourceFile elems) = ppElems elems

ppElems NoElem = empty
ppElems (OneElem elem elems _) = ppElem elem >-< ppElems elems

ppElem (DefElem def) = ppDef def >|< text ";"
ppElem (TrafoElem trafo) = ppTransform trafo

ppProg (Prog b) = ppBlock b

ppBlock Empty                  = empty
ppBlock (DefsBlock dlist block) = wrapbraces (ppDlist dlist) 
                                 >-< (pp_sp >-< ppBlock block)
ppBlock (TrafoBlock trafo block) = ppTransform trafo
                                 >-< (pp_sp >-< ppBlock block)

-- ugh!

theorytorawtheor (Theory theo) = theotorawtheor theo
theotorawtheor NoTheo = []
theotorawtheor (CeqTheo ceq theo) = Left (ceqtoceq ceq):theotorawtheor theo
theotorawtheor (DefnTheo name theo) = Right name:theotorawtheor theo

theorytotheor (Theory theo) = theototheor theo
theototheor NoTheo = []
theototheor (CeqTheo ceq theo) = ceqtoceq ceq:theototheor theo
theototheor (DefnTheo name theo) 
 = error "Unexpected definition found in post-typechecking theory"

ceqtoceq (Ceq name conds eqn _) = (name,eqstoeqns conds,eqntoeqn eqn)

eqstoeqns NoEq = []
eqstoeqns (OneEq eqn eqs) = eqntoeqn eqn:eqstoeqns eqs

eqntoeqn (Eqn lhs rhs) = (lhs,rhs)

ppTransform (Rewrite name depends redef spec using)
 =     text "TRANSFORM " >|< text name >|< printdeps depends
   >-< text "REDEFINE " >|< ppDlist redef
   >-< text "SPECIFYING " >|< ppDlist spec
   >-< text "USING " >|< ppRawTheor (theorytorawtheor using)
   >-< text "END;"
          where printdeps [] = empty
                printdeps xs = text ":" >|< printdeps' xs
                printdeps' [x] = text x
                printdeps' (x:xs) = text x >|< text "," >|< printdeps' xs

wrapbraces ds = vlist (text "{" : ds ++ [text "}"])

ppDlist (DList d)
          = initmap newline (map ppDef (dlistify d))
            where dlistify NoDef         = []
                  dlistify (OneDef d ds _) = d:dlistify ds

newline  y = y >|< text ";"
nnewline y = y >|< text ";\n"

initmap f []    = []
initmap f [a]   = [a]
initmap f (a:x) = f a : initmap f x

ppDef (Fun lhs rhs)
  = (eelement_h1 par >>|<< par  >>^<<  par >>-<< eindent 3 par) >>$< [plhs, prhs]
  where plhs = ppExp 0 lhs
        prhs = text " = " >|< ppExp 0 rhs
ppDef (Data t as)   = text "data " >|< ppType t    >|< ppAlts as
ppDef (Dec  e t )   = ppExp 0 e    >|< text " :: " >|< ppType t

ppAlts as        = hv_sep " = " " | " . cata_Alts ((:) . ppAlt ,[]) $ as
ppAlt (Alt n ts) = ppType (TCon n ts)

cata_Alts (op,e) = cata
  where cata as = case as of
                    NoAlt       -> e
                    OneAlt a as -> op a (cata as)

ppExp i (Lam p e _) 
   = ppLam [p] e
     where ppLam ps (Lam p e _)  = ppLam (ps++[p]) e
           ppLam ps e
             = hv [ "(" >|< ppHead (text "\\") ps
                  , pp_sp >|< ppExp 0 e
                  , pp ")"
                  ]
           ppHead  ppArgs (p:ps) = ppHead' (ppArgs >#< ppExp (prec App) p) ps
           ppHead' ppArgs []     = ppArgs >|< text " ->"
           ppHead' ppArgs (p:ps)
             = ppHead' (case p
                        of Var v _ -> ppArgs >#< text v
                           _       -> ppArgs >-< element_h1 (ppExp (prec App) p)) ps

ppExp i (Let ds e _) = paren i 1 (ppExp 1 e >-< 
                                  (text "where " >|< wrap (ppDlist ds)))
                       where wrap []  = empty
                             wrap [x] = x
                             wrap xs  = vlist (text "{" : xs ++ [text "}"])

ppExp i (Ap fun arg _) 
   = ppAp [arg] fun
     where ppAp as (Ap f a _) = ppAp (a:as) f
           ppAp as (Var x _)  = ppAps i (text x) as 
           ppAp as (Con c _)  = ppCon i c as
           ppAp as l@(Lam _ _ _)
             = ppAps i (ppExp 0 l) as
           ppAp as e          = ppAps i (pp_parens (ppExp 0 e)) as
ppExp i (Var x _) = text x
ppExp i (Con c _) = ppCon i c []

allTypes e = prettyExp e ++ "::" ++ prettyType (exptype e) ++ "\n" ++
             subtypes e
subtypes (Ap e1 e2 _) = allTypes e1 ++ allTypes e2
subtypes (Lam p e _)  = allTypes p ++ allTypes e
subtypes _            = []

ppAps i x as = if null as then x
                          else paren i (prec App) (x >#< hv_sp ppas)
  where ppas = map (ppExp (prec App)) as

paren i j e = if j<=i then pp_parens e else e

ppCon i (Number n)  as = ppAps   i (text (show n)) as
ppCon i (FNumber n) as = ppAps   i (text (show n)) as
ppCon i (Boolean b) as = ppAps   i (text (show b)) as
ppCon i (Binop b)   as = ppBinop i b               as
ppCon i If          as = ppIf    i                 as 
ppCon i Nul         as = ppAps   i (text "[]")     as
ppCon i (Tuple n)   as = ppTuple n                 as
ppCon i (Constr x)  as = ppAps   i (text x)        as
ppCon i Error       as = ppAps   i (text "Error")  as

ppIf i (c:t:e:as) = if null as then paren i 1 ifonly
                               else ppAps i (pp_parens ifonly) as
  where ifonly = pp_ite "if " " then " " else " "" (ppExp 0 c) (ppExp 0 t) (ppExp 0 e)
ppIf i as         = ppAps i (text "If") as

ppBinop i op [a,b]    = paren i j (ppBin j a op b)
  where j = prec op
ppBinop i op (a:b:as) = ppAps i (pp_parens (ppBin j a op b)) as
  where j = prec op
ppBinop i op as       = ppAps i (pp_parens (text (opText op))) as

ppBin j (Var a _) op (Var b _) = hlist_sp [pp a, ppop, pp b]
  where ppop = text (opText op)
ppBin j a op b = hv_sp [ppa,ppop,ppb]
  where ppa  = ppExp j a
        ppop = text (opText op) 
        ppb  = ppExp j b
                  
ppTuple n as = if length as >= n 
               then foldr1 (>#<) 
                     (pp_block "(" ")" ", " (map (ppExp 0) bs) : map (ppExp (prec App)) cs) 
               else hlist_sp (text ("Tuple " ++ show n) : map (ppExp (prec App)) as)
  where (bs,cs) = splitAt n as

opT op = pp_sp >|< text (opText op) >|< pp_sp

opText App = " "
opText Idx = "!"
opText Exp = "^"
opText FExp = "^."
opText Comp = "."
opText Tim = "*"
opText Div = "/"
opText Mod = "mod"
opText Plus = "+"
opText Min = "-"
opText FTim = "*."
opText FDiv = "/."
opText FPlus = "+."
opText FMin = "-."
opText Cat = "++"
opText Cons = ":"
opText Eq = "=="
opText Neq = "/="
opText Le = "<"
opText Leq = "<="
opText Ge = ">"
opText Geq = ">="
opText And = "&&"
opText Or = "||"

ppType (TCon "Arrow" [s,t])              = hv_sp [ppAtom s, text "->", ppType t]
ppType (TCon "List" [t])                 = pp_brackets (ppType t)
ppType (TCon x ts) | take 5 x == "Tuple" = pp_parens (psepBy ',' (map ppType ts))
                   | otherwise           = psepBy ' ' (text x:map ppAtom ts)
ppType (Tid x    )                       = text x
ppType (TAbs xs t)                       = ppType t

psepBy c = foldr1 step 
           where step x ys = x >|< pp_char c >|< ys

ppAtom (Tid x)             = text x
ppAtom t@(TCon "Arrow" ts) = pp_parens (ppType t)
ppAtom t                   = if atomic t then ppType t
                                         else pp_parens ( ppType t )

atomic (TCon "List" _)      = True
atomic (TCon "Arrow" [s,t]) = False
atomic (TCon x [])          = True
atomic _                    = False

ppTheor = vlist . initmap nnewline . map ppCeq

ppRawTheor = vlist . initmap nnewline . map (either ppCeq ppDefnRule)

ppDefnRule name = hv [text "DEFINITION " >|< text name]

ppCeq (n,cs,e) =  hv [text n >|< text ": ", ppEqn e, ppConds cs]

ppEqn (e1,e2) = ppDef (Fun e1 e2) 

ppConds [] = empty
ppConds es = text ", if " 
             >|< vlist (text "{" : initmap newline (map ppEqn es) ++ [text "}"])

