module State(StT(..),liftSt,evalSt,readSt,writeSt,applSt,
               transformStT,catchSt,embedIOSt,
               St(..),appl,eval,readS,writeS)
where
import Monad
import IOExts -- needed for embedIOSt below

-- State Transformer

newtype StT a m b = StT (a -> m (b, a))

instance Monad m => Monad (StT a m) where
    return a = liftSt (return a)
    StT x >>= f = StT (\s -> x s >>= \(v,s') -> applSt (f v) s')
    fail str = liftSt (fail str)

applSt :: Monad m => StT a m b -> a -> m (b,a)
applSt (StT g) y = g y

liftSt :: Monad m => m b -> StT a m b
liftSt m = StT (\s -> m >>= \v -> return (v,s))

readSt :: Monad m => StT a m a
readSt = StT f
       where f s = return (s,s)

writeSt :: Monad m => a -> StT a m ()
writeSt s = StT f 
       where f _ = return ((),s)

transformStT :: (Monad m,Monad n) => (m (b,a) -> n (b,a)) 
                                       -> StT a m b -> StT a n b
transformStT tr (StT f) = StT (tr.f)

catchSt :: StT a IO b -> (IOError -> StT a IO b) -> StT a IO b
catchSt (StT f) fn = StT g
           where g s = catch (f s) fn'
                    where   fn' error = do let StT h = fn error
                                           h s


evalSt :: Monad m => a -> StT a m b -> m b
evalSt a (StT f) = do (x,s) <- f a
                      return x

instance Monad m => Functor (StT a m) where
    fmap f (StT x) = StT (\s -> do (b,s') <- x s
                                   return (f b,s'))

instance MonadPlus m => MonadPlus (StT a m) where
    mzero = liftSt mzero
    (StT x) `mplus` (StT y) = StT (\s -> x s `mplus` y s)

-- useful for debugging StT a IO b via the ghci interactive loop

embedIOSt :: IORef a -> StT a IO b -> IO b
embedIOSt ref (StT f) = do s <- readIORef ref
                           (b,s') <- f s
                           writeIORef ref s'
                           return b

-- newtype Id a = Id a
-- instance Monad Id where
--   return x = Id x
--   Id x >>= f = f x


-- State monad

newtype St a b = St (a -> (b, a))

instance Monad (St a) where
    return a = St (\ xs -> (a,xs))
    p >>= q = St f
              where f s = appl (q x) s'
                          where (x,s') = appl p s 

appl :: St a b -> a -> (b,a)
appl (St f) x = f x

eval :: a -> St a b -> b
eval a (St f) = fst (f a)

readS :: St a a
readS = St (\a -> (a,a))

writeS :: a -> St a ()
writeS a = St (\_ -> ((),a))
